//
//  SignNUC.m
//  CESNA
//
//  Created by Alisher Makulbayev on 29.07.16.
//
//

#import "SignNUC.h"
//#import "AppDelegate.h"

#if EDC_WORK

#import "NSData+Base64.h"
#include "libxml/parser.h"
#include "libxml/tree.h"
#include "libxml/c14n.h"
#include "libxml/xpath.h"
#include "libxml/xpathInternals.h"
#include "openssl/engine.h"
#include "openssl/pkcs12.h"

#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/evp.h"
#include "openssl/conf.h"
#include "openssl/pem.h"
#include "openssl/comp.h"
#include "openssl/x509v3.h"
#include "stdio.h"
#include "obj_mac.h"
#define CALG_SHA256   32780

const char* NS_XMLDSIG = "http://www.w3.org/2000/09/xmldsig#";
const char* C14N_OMIT_COMMENTS = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315";
const char* C14N_WITH_COMMENTS = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments";
const char* ALG_GOST34310 = "http://www.w3.org/2001/04/xmldsig-more#gost34310-gost34311";
const char* ALG_TRANSFORM = "http://www.w3.org/2000/09/xmldsig#enveloped-signature";
const char* ALG_GOST3411 = "http://www.w3.org/2001/04/xmldsig-more#gost34311";
const char* ALG_RSA = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256";
const char* ALG_SHA256 = "http://www.w3.org/2001/04/xmlenc#sha256";

#endif

@implementation SignNUC

int GetHash2 (unsigned char * data, unsigned long data_len,
              unsigned char * md_value, unsigned int * md_len)
{
#if (EDC_WORK)
    EVP_MD_CTX *mdctx;
    const EVP_MD *md;
    md = EVP_sha256();
    if(!md) return 1;
    
    mdctx = EVP_MD_CTX_create();
    EVP_DigestInit_ex(mdctx, md, NULL);
    
    EVP_DigestUpdate(mdctx, data, data_len);
    EVP_DigestFinal_ex(mdctx, md_value, md_len);
    EVP_MD_CTX_destroy(mdctx);
#endif

    return 0;
}
 
int SignHashWithPKCS12 (
                       const char* data, unsigned long data_length,
                       unsigned char* nameKey, const char * pin,
                       unsigned char* signData, unsigned long* signLength,
                       unsigned char* certData, unsigned int* certLength,
                       int isRSA,
                       unsigned int hashAlg)
{
#if (EDC_WORK)
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *basePath = ([paths count] > 0) ? [paths objectAtIndex:0] : nil;
    
    NSString *pfile =[basePath stringByAppendingFormat:@"/%s.%@",nameKey,@"p12"];
    
    FILE *fp;
    PKCS12 *p12;
    EVP_PKEY *pkey;
    X509 *cert;
    int err;
    unsigned long sig_len;
    
    if(!data)
        return -1;
    
    BIO *databio = BIO_new(BIO_s_mem());
    BIO_puts(databio, data);
    
    STACK_OF(X509) *ca = NULL;
    
    if(![[NSFileManager defaultManager] fileExistsAtPath:pfile]) return -2;
    
    
    fp = fopen([pfile UTF8String], "rb");
    p12 = d2i_PKCS12_fp(fp, NULL);
    fclose (fp);
    
    if (!p12) {
        ERR_print_errors_fp(stderr);
        return -3;
    }
    
    if (!PKCS12_parse(p12, pin, &pkey, &cert, &ca)) {
        ERR_print_errors_fp(stderr);
        return -4;
    }
    
    *certLength = i2d_X509(cert, &certData);
    
    PKCS12_free(p12);
    
    EVP_MD_CTX md_ctx;
    const EVP_MD *md = NULL;
    int algnid = pkey->type;
    
    if(algnid == NID_id_GostOld34310_2004) {
        md = EVP_get_digestbynid(NID_id_GostOld34311_95); //64
        sig_len = 64;
    } else if(algnid == NID_id_Gost34310_2004) {
        md = EVP_get_digestbynid(NID_id_Gost34311_95);
        sig_len = 64;
    } else {
        md = EVP_sha256();
        sig_len = 256;
    }
    
    unsigned char sig_buf [sig_len];
    ERR_print_errors_fp(stderr);
    
    if(!md) {
        ERR_print_errors_fp(stderr);
        return -5;
    }
    
    EVP_SignInit(&md_ctx, md);
    EVP_SignUpdate (&md_ctx, data, data_length);
    sig_len = sizeof(sig_buf);
    err = EVP_SignFinal (&md_ctx, signData,(unsigned int *)&sig_len, pkey);
    
    if (err != 1) {
        ERR_print_errors_fp(stderr);
        return -6;
    }
    *signLength = sig_len;
#endif 
    
    return 0;
}



+(int)xmlSigner:(NSString**)signXML withDATA:(NSString*)dataToSign CertName:(NSString*)certName andPassword:(NSString*)password
{
#if (EDC_WORK)
    OpenSSL_add_all_algorithms();
    ENGINE_load_openssl();
    
    xmlInitParser();
    LIBXML_TEST_VERSION;
    
    xmlDocPtr doc = NULL;
    xmlNodePtr
    root = NULL,
    signEl = NULL,
    sInfoEl = NULL,
    canMethEl = NULL,
    signMethEl = NULL,
    refEl = NULL,
    transEl = NULL,
    tranEl = NULL,
    tran2El = NULL,
    digMethEl = NULL,
    digValEl = NULL,
    sigValEl = NULL,
    kInfoEl = NULL,
    x509DataEl = NULL,
    x509CertEl = NULL;
    
    unsigned int isRSA = 1;
    unsigned int hashSize = 20;
    unsigned long signLength = 256;
    unsigned int hashAlg = CALG_SHA256;
    xmlChar *xmlHashAlg = BAD_CAST ALG_SHA256;
    xmlChar *xmlSignAlg = BAD_CAST ALG_RSA;
    
    if (isRSA == 0) {
        hashSize = 32;
        signLength = 64;
        hashAlg = 0;
        xmlSignAlg = BAD_CAST ALG_GOST34310;
        xmlHashAlg = BAD_CAST ALG_GOST3411;
    }
    
    NSData *xml = [dataToSign dataUsingEncoding:NSUTF8StringEncoding];
    if(!xml) {
        NSLog(@"Xml was not loaded");
        return 1;
    }
    
    unsigned char* cXml = (unsigned char*)malloc(xml.length);
    [xml getBytes:cXml length:xml.length];
    cXml[xml.length] = 0x0;
    
    doc = xmlParseDoc(cXml);
    xmlChar* c14nXML = NULL;
    xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &c14nXML);
    
    
    unsigned long c14nXMLLen = strlen((char*)c14nXML);
    unsigned char* cHash = (unsigned char*)malloc(10000);
    
    int res = GetHash2(c14nXML, c14nXMLLen, cHash, &hashSize);
    
    NSData *hashData = [NSData dataWithBytes:cHash length:hashSize];
    if (res != 0) return 2; //GetHash2 fail
    
    xmlXPathContextPtr xpathCtx;
    xmlXPathObjectPtr xpathObj;
    xmlNodeSetPtr sInfoNS;
    
    root = xmlDocGetRootElement(doc);
    signEl = xmlNewNode(NULL, BAD_CAST "ds:Signature");
    xmlNsPtr signNS = xmlNewNs(signEl, BAD_CAST NS_XMLDSIG, BAD_CAST "ds");
    xmlAddChild(root, signEl);
    sInfoEl = xmlNewChild(signEl, signNS, BAD_CAST "SignedInfo", NULL);
    canMethEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "CanonicalizationMethod", NULL);
    xmlNewProp(canMethEl, BAD_CAST "Algorithm", BAD_CAST C14N_OMIT_COMMENTS);
    signMethEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "SignatureMethod", NULL);
    xmlNewProp(signMethEl, BAD_CAST "Algorithm", xmlSignAlg);
    refEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "Reference", NULL);
    xmlNewProp(refEl, BAD_CAST "URI", NULL);
    transEl = xmlNewChild(refEl, signNS, BAD_CAST "Transforms", NULL);
    tranEl = xmlNewChild(transEl, signNS, BAD_CAST "Transform", NULL);
    xmlNewProp(tranEl, BAD_CAST "Algorithm", BAD_CAST ALG_TRANSFORM);
    tran2El = xmlNewChild(transEl, signNS, BAD_CAST "Transform", NULL);
    xmlNewProp(tran2El, BAD_CAST "Algorithm", BAD_CAST C14N_WITH_COMMENTS);
    digMethEl = xmlNewChild(refEl, signNS, BAD_CAST "DigestMethod", NULL);
    xmlNewProp(digMethEl, BAD_CAST "Algorithm", xmlHashAlg);
    xmlChar *b64DigVal = (xmlChar*)[[hashData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    digValEl = xmlNewChild(refEl, signNS, BAD_CAST "DigestValue", BAD_CAST b64DigVal);
    
    xpathCtx = xmlXPathNewContext(doc);
    xmlXPathRegisterNs(xpathCtx, BAD_CAST "ds", BAD_CAST NS_XMLDSIG);
    xpathObj = xmlXPathEvalExpression(BAD_CAST "(//. | //@* | //namespace::*)[ancestor-or-self::ds:SignedInfo]", xpathCtx);
    sInfoNS = xpathObj->nodesetval;
    
    xmlChar *c14nSInfo = NULL;
    xmlC14NDocDumpMemory(doc, sInfoNS, 0, NULL, 1, &c14nSInfo);
    xmlXPathFreeObject(xpathObj);
    xmlXPathFreeContext(xpathCtx);
    
    unsigned long c14nSInfoLen = strlen((char*)c14nSInfo);
    res = GetHash2(c14nSInfo, c14nSInfoLen, cHash, &hashSize);
    
    hashData = [NSData dataWithBytes:cHash length:hashSize];
    if (res != 0)
        return 3; //GetHash2 fail
    
    
    NSData *signedData = NULL;
    unsigned char* cSignData = (unsigned char*)malloc(10000);
    
    NSData *certData = NULL;
    unsigned char* cCertData = (unsigned char*)malloc(10000);
    unsigned int certLength = 0;
    
    res = SignHashWithPKCS12((const char*)cHash, hashSize, (unsigned char*)[certName UTF8String], [password UTF8String], cSignData, &signLength, cCertData, &certLength, isRSA, hashAlg);
    
    cSignData[signLength] = 0x0;
    if (res != 0)
        return res; //ERROR sign
    
    signedData = [NSData dataWithBytes:cSignData length:signLength];
    certData = [NSData dataWithBytes:cCertData length:certLength];
    
    xmlChar *b64SigVal = (xmlChar*)[[signedData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    sigValEl = xmlNewChild(signEl, signNS, BAD_CAST "SignatureValue", BAD_CAST b64SigVal);
    kInfoEl = xmlNewChild(signEl, signNS, BAD_CAST "KeyInfo", NULL);
    x509DataEl = xmlNewChild(kInfoEl, signNS, BAD_CAST "X509Data", NULL);
    xmlChar *b64CertVal = (xmlChar*)[[certData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    x509CertEl = xmlNewChild(x509DataEl, signNS, BAD_CAST "X509Certificate", BAD_CAST b64CertVal);
    
    xmlChar *outXML = NULL;
    int outXMLSize = 0;
    xmlDocDumpMemoryEnc(doc, &outXML, &outXMLSize, "UTF-8");
    *signXML = [NSString stringWithCString:(const char*)outXML encoding:NSUTF8StringEncoding];
    printf("\nxml =\n%s", outXML);
    
    xmlFreeDoc(doc);
    xmlCleanupParser();
    xmlMemoryDump();
    
    free(cCertData);
    free(cHash);
    free(outXML);
    free(cSignData);
    ERR_free_strings();
#endif
    
    return 0;
}


@end
