//
//  XMLSigner.m
//  SSLTESTER
//
//  Created by Dev on 06/02/15.
//  Copyright (c) 2015 STS. All rights reserved.
//

#import <Foundation/Foundation.h>
#import "XMLSigner.h"
#include <stdio.h>
#include "openssl/engine.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/evp.h"
#include "openssl/conf.h"
#include "openssl/pkcs12.h"
#include "openssl/pem.h"
#include "openssl/comp.h"
#include "openssl/x509.h"
#include "openssl/x509v3.h"
#include "obj_mac.h"
#define CALG_SHA1   32772

const char* NS_XMLDSIG = "http://www.w3.org/2000/09/xmldsig#";
const char* C14N_OMIT_COMMENTS = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315";
const char* C14N_WITH_COMMENTS = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments";
const char* ALG_GOST34310 = "http://www.w3.org/2001/04/xmldsig-more#gost34310-gost34311";
const char* ALG_TRANSFORM = "http://www.w3.org/2000/09/xmldsig#enveloped-signature";
const char* ALG_GOST3411 = "http://www.w3.org/2001/04/xmldsig-more#gost34311";
const char* ALG_RSA = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha1";
const char* ALG_SHA1 = "http://www.w3.org/2001/04/xmldsig-more#sha1";

int GetHash2(unsigned char * data, int data_len, unsigned char * md_value, int md_len, int isRSA) {
    EVP_MD_CTX *mdctx; 
    const EVP_MD *md;
    int i;

    if(isRSA == 0)
        md = EVP_sha1();// TODO: Dont use sha1!!! Its deprecated. Use sha2....
    else
        md = EVP_get_digestbyname("md_gost95");
    
    if(!md) {
        NSLog(@"Unknown message digest md_gost95\n");
        return 1;
    }
    
    mdctx = EVP_MD_CTX_create();
    EVP_DigestInit_ex(mdctx, md, NULL);

    EVP_DigestUpdate(mdctx, data, data_len);
    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    EVP_MD_CTX_destroy(mdctx);
    
    printf("Digest is: ");
    for(i = 0; i < md_len; i++) {
        printf("%02x", md_value[i]);
    }
    printf("\n");
    
//    EVP_cleanup();
    
    return 0;
}

int SignHashWithPKCS12(unsigned char* data, unsigned int data_length, unsigned char* pkcs12_path, unsigned char * pin, unsigned char* signData, unsigned int* signLength, unsigned char* certData, unsigned int* certLength, int isRSA, unsigned int hashAlg) {
    
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *basePath = ([paths count] > 0) ? [paths objectAtIndex:0] : nil;
    
    NSString *outf = [basePath stringByAppendingString:@"/outxml.pem"];
    NSString *signf = [basePath stringByAppendingString:@"/signxml"];
    
    NSString *pfile = [NSString stringWithUTF8String:pkcs12_path];
    FILE *fp;
    PKCS12 *p12;
    EVP_PKEY *pkey;
    X509 *cert;
    int err;
    int sig_len;
    
    if(!data)
        return 1;

    BIO *databio = BIO_new(BIO_s_mem());
    BIO_puts(databio, data);
    
    STACK_OF(X509) *ca = NULL;
    
    if([[NSFileManager defaultManager] fileExistsAtPath:pfile]) {
        NSLog(@"ok, pfile exists!");
    } else {
        NSLog(@"error, pfile does not exists!");
        return 1;
    }
    
    fp = fopen([pfile UTF8String], "rb");
    p12 = d2i_PKCS12_fp(fp, NULL);
    fclose (fp);
    if (!p12) {
        fprintf(stderr, "Error reading PKCS#12 file\n");
        ERR_print_errors_fp(stderr);
        return 1;
    }
    
    if (!PKCS12_parse(p12, pin, &pkey, &cert, &ca)) { //Error at parsing or pin error
        fprintf(stderr, "Error parsing PKCS#12 file\n");
        ERR_print_errors_fp(stderr);
        return 1;
    }

    PKCS12_free(p12);
    
    EVP_MD_CTX md_ctx;
    const EVP_MD *md = NULL;
    //Algorithm
    int algnid = pkey->type;
    if(algnid == NID_id_GostOld34310_2004) {
        md = EVP_get_digestbynid(NID_id_GostOld34311_95); //64
        sig_len = 64;
    } else if(algnid == NID_id_Gost34310_2004) {
        md = EVP_get_digestbynid(NID_id_Gost34311_95);
        sig_len = 64;
    } else {
        //SOON, extra algs and other lengs
        md = EVP_sha1();// TODO: Dont use sha1!!! Its deprecated. Use sha2....
        sig_len = 256;
    }
    
    unsigned char sig_buf [sig_len];
    ERR_print_errors_fp(stderr);
    
    if(!md) {
        fprintf(stderr, "Error creating message digest");
        ERR_print_errors(stderr);
        return 1;
    }
    data[data_length] = 0x0;
    //Singing Выберите то что вам нужно из двух видов.
    EVP_SignInit(&md_ctx, md);
    EVP_SignUpdate (&md_ctx, data, data_length);
    sig_len = sizeof(sig_buf);
    err = EVP_SignFinal (&md_ctx, signData, &sig_len, pkey);

    if (err != 1) {
        ERR_print_errors_fp(stderr);
        return 1;
    }
    *signLength = sig_len;
//    NSData* sign = [NSData dataWithBytes:(const void *)signData length:sig_len];
//    [sign writeToFile:signf atomically:NO];
    return 0;
}

BOOL xmlsignersign() {
//    OPENSSL_config(NULL);
    OpenSSL_add_all_algorithms();
    
//    ENGINE_load_openssl();
    ENGINE_load_gost();
    
    xmlInitParser();
    LIBXML_TEST_VERSION;
    
    xmlDocPtr doc = NULL;
    xmlNodePtr root = NULL, signEl = NULL, sInfoEl = NULL, canMethEl = NULL, signMethEl = NULL, refEl = NULL, transEl = NULL, tranEl = NULL, tran2El = NULL, digMethEl = NULL, digValEl = NULL, sigValEl = NULL, kInfoEl = NULL, x509DataEl = NULL, x509CertEl = NULL;
    
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *basePath = ([paths count] > 0) ? [paths objectAtIndex:0] : nil;
    
    unsigned int isRSA = 0; // 0 - gost, 1 - rsa
    unsigned int hashSize = 20;
    unsigned int signLength = 256;
    unsigned int hashAlg = CALG_SHA1;
    xmlChar *xmlHashAlg = BAD_CAST ALG_SHA1;
    xmlChar *xmlSignAlg = BAD_CAST ALG_RSA;
    
    NSData *blob, *cert;
    
    if (isRSA == 0) {
        hashSize = 32;
        signLength = 64;
        hashAlg = 0;
        xmlSignAlg = BAD_CAST ALG_GOST34310;
        xmlHashAlg = BAD_CAST ALG_GOST3411;
    }
    NSData *pkcs12;
    pkcs12 = [NSData dataWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"RSA_d215fe883a5317a3784808594d3ed393140476e4" ofType:@"p12"]];
    unsigned char* cPkcs12 = (unsigned char*)malloc(pkcs12.length);
    [pkcs12 getBytes:cPkcs12 length:pkcs12.length];
    
    NSData *xml = [NSData dataWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"my" ofType:@"xml"]];

    //NSString *xmlString  = @"<root>test</root>";
    //xml = [xmlString dataUsingEncoding:NSUTF8StringEncoding];
    if(!xml) {
        NSLog(@"Xml was not loaded");
        return 1;
    }
    
    unsigned char* cXml = (unsigned char*)malloc(xml.length);
    [xml getBytes:cXml length:xml.length];
    
    //    cXml = BAD_CAST "<p:requestIin xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xmlns:p=\"http://message.persistence.interactive.nat\" xsi:type=\"p:RequestIin\"><GUID>552CB4E9-013A-4000-E000-59870A0C0A02</GUID><dateMessage>2012-10-12T13:32:04.201Z</dateMessage><senderCode>PORTAL</senderCode><iin>800513300402</iin></p:requestIin>";
    cXml[xml.length] = 0x0;

    doc = xmlParseDoc(cXml);
    xmlChar* c14nXML = NULL;
    xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &c14nXML);
    int c14nXMLLen = strlen((char*)c14nXML);

    unsigned char* cHash = (unsigned char*)malloc(10000);
    
    int res = GetHash2(c14nXML, c14nXMLLen, cHash, &hashSize, isRSA);
    
    NSData *hashData = [NSData dataWithBytes:cHash length:hashSize];
    if (res == 0) {
        NSLog(@"Canonicalized source XML has been hashed");
        NSString *hashPath = [basePath  stringByAppendingString:@"/c14nSrcHash.bin"];
        BOOL hb = [hashData writeToFile:hashPath atomically:NO];
        if (hb) {
            NSLog(@"path to hash: %@", hashPath);
        }
    } else {
        NSLog(@"GetHash2 fail");
        return 1;
    }
    
    xmlXPathContextPtr xpathCtx;
    xmlXPathObjectPtr xpathObj;
    xmlNodeSetPtr sInfoNS;
    
    root = xmlDocGetRootElement(doc);
    signEl = xmlNewNode(NULL, BAD_CAST "ds:Signature");
    xmlNsPtr signNS = xmlNewNs(signEl, BAD_CAST NS_XMLDSIG, BAD_CAST "ds");
    xmlAddChild(root, signEl);
    sInfoEl = xmlNewChild(signEl, signNS, BAD_CAST "SignedInfo", NULL);
    canMethEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "CanonicalizationMethod", NULL);
    xmlNewProp(canMethEl, BAD_CAST "Algorithm", BAD_CAST C14N_OMIT_COMMENTS);
    signMethEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "SignatureMethod", NULL);
    xmlNewProp(signMethEl, BAD_CAST "Algorithm", xmlSignAlg);
    refEl = xmlNewChild(sInfoEl, signNS, BAD_CAST "Reference", NULL);
    xmlNewProp(refEl, BAD_CAST "URI", NULL);
    transEl = xmlNewChild(refEl, signNS, BAD_CAST "Transforms", NULL);
    tranEl = xmlNewChild(transEl, signNS, BAD_CAST "Transform", NULL);
    xmlNewProp(tranEl, BAD_CAST "Algorithm", BAD_CAST ALG_TRANSFORM);
    tran2El = xmlNewChild(transEl, signNS, BAD_CAST "Transform", NULL);
    xmlNewProp(tran2El, BAD_CAST "Algorithm", BAD_CAST C14N_WITH_COMMENTS);
    digMethEl = xmlNewChild(refEl, signNS, BAD_CAST "DigestMethod", NULL);
    xmlNewProp(digMethEl, BAD_CAST "Algorithm", xmlHashAlg);
    xmlChar *b64DigVal = (xmlChar*)[[hashData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    digValEl = xmlNewChild(refEl, signNS, BAD_CAST "DigestValue", BAD_CAST b64DigVal);
    
    xpathCtx = xmlXPathNewContext(doc);
    xmlXPathRegisterNs(xpathCtx, BAD_CAST "ds", BAD_CAST NS_XMLDSIG);
    xpathObj = xmlXPathEvalExpression(BAD_CAST "(//. | //@* | //namespace::*)[ancestor-or-self::ds:SignedInfo]", xpathCtx);
    sInfoNS = xpathObj->nodesetval;
    
    xmlChar *c14nSInfo = NULL;
    xmlC14NDocDumpMemory(doc, sInfoNS, 0, NULL, 1, &c14nSInfo);
    xmlXPathFreeObject(xpathObj);
    xmlXPathFreeContext(xpathCtx);
    
    int c14nSInfoLen = strlen((char*)c14nSInfo);
    res = GetHash2(c14nSInfo, c14nSInfoLen, cHash, &hashSize, isRSA);
    
    hashData = [NSData dataWithBytes:cHash length:hashSize];
    if (res == 0) {
        NSLog(@"Canonicalized SignedInfo has been hashed");
        NSString *hashPath = [basePath  stringByAppendingString:@"/c14nHash.bin"];
        BOOL hb = [hashData writeToFile:hashPath atomically:NO];
        if (hb) {
            NSLog(@"path to hash: %@", hashPath);
        }
    } else {
        NSLog(@"GetHash2 Fail");
        return 1;
    }
    
    NSData *signedData = NULL;
    unsigned char* cCertData = (unsigned char*)malloc(10000);
    unsigned int certLength = 0;
    unsigned char* cSignData = (unsigned char*)malloc(10000);
    
    NSData *certData = NULL;

    NSString *path = [[NSBundle mainBundle] pathForResource:@"RSA_d215fe883a5317a3784808594d3ed393140476e4" ofType:@"p12"];
    @try {
            res = SignHashWithPKCS12(cHash, hashSize, [path UTF8String], (unsigned char *)"123456", cSignData, &signLength, cCertData, &certLength, isRSA, hashAlg);
        }
        @catch (NSException *exception) {
                 signedData = [NSData dataWithBytes:cSignData length:signLength];
                 NSString *signPath = [basePath stringByAppendingString:@"/signedHash.bin"];
                 BOOL sb = [signedData writeToFile:signPath atomically:NO];
        }
    //TODO : To be or not to be?! Need or make this part as the option?
    //Needs to reverse sign
    unsigned char *test = (unsigned char*)malloc(signLength);
    int j;
    for(j =0; j < signLength; j++) {
        test[j] = cSignData[signLength - j - 1];
    }
    for(j =0; j < signLength; j++) {
        cSignData[j] = test[j];
    }
    free(test);
    
    cSignData[signLength] = 0x0;
    if (res == 0) {
        NSLog(@"SignHashWithPKCS12 success");
        signedData = [NSData dataWithBytes:cSignData length:signLength];
        NSString *signPath = [basePath stringByAppendingString:@"/signedHash.bin"];
        BOOL sb = [signedData writeToFile:signPath atomically:NO];
        if (sb) {
            NSLog(@"path to signedHash: %@", signPath);
        }
        certData = [NSData dataWithBytes:cCertData length:certLength];
    } else {
        NSLog(@"SignHashWithPKCS12 Fail");
        return 1;
    }
    
    xmlChar *b64SigVal = (xmlChar*)[[signedData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    sigValEl = xmlNewChild(signEl, signNS, BAD_CAST "SignatureValue", BAD_CAST b64SigVal);
    kInfoEl = xmlNewChild(signEl, signNS, BAD_CAST "KeyInfo", NULL);
    x509DataEl = xmlNewChild(kInfoEl, signNS, BAD_CAST "X509Data", NULL);
    xmlChar *b64CertVal = (xmlChar*)[[certData base64EncodedString] cStringUsingEncoding:[NSString defaultCStringEncoding]];
    x509CertEl = xmlNewChild(x509DataEl, signNS, BAD_CAST "X509Certificate", BAD_CAST b64CertVal);
    
    xmlChar *outXML = NULL;
    int outXMLSize = 0;
    xmlDocDumpMemoryEnc(doc, &outXML, &outXMLSize, "UTF-8");
    NSData *signedXML = [NSData dataWithBytes:outXML length:outXMLSize];
    NSString *signedXMLPath = [basePath stringByAppendingString:@"/signedXML.xml"];
    BOOL sb = [signedXML writeToFile:signedXMLPath atomically:NO];
    if (sb) {
        NSLog(@"path to signedXML: %@", signedXMLPath);
    }
    printf("\nxml =\n%s",outXML);

    xmlFreeDoc(doc);
    xmlCleanupParser();
    xmlMemoryDump();
    
    free(cCertData);
    free(cPkcs12);
    free(cHash);
    free(outXML);
    free(cSignData);
//    EVP_cleanup();    // If call this one, it cleans all digests and ciphers. So dont call it if you didnt end work
//    return YES;
//    ENGINE_cleanup();
    ERR_free_strings();

    return 0;
}
